#ifndef __CHART__
#define __CHART__

#include <iostream>
#include <fstream>
#include <vector>

#include "util.h"

namespace SD {


struct Cell {
    Label label;
    Cell* next;
    Cell* prev;
    Real  score;
    Real  alpha;
    Real  beta;
    Real  lb;
    Cell* backlink;
};


class Chart {
  private:
    static const Real EPSILON;
    uint  L_;
    uint  D_;
    Cell* cell_;
    std::vector<Cell*> first_;
    std::vector<Cell*> last_;
  public:
    Chart();
    Chart( const Chart& obj );
    ~Chart();
    Chart& operator=( const Chart& obj );
    void set_size( uint L, uint D );
    void viterbi_init( uint N );
    void staggered_init( uint N );
    
    Cell* first( uint n );
    const Cell* first( uint n ) const;
    Cell* last( uint n );
    const Cell* last( uint n ) const;
    void expand( uint n );
    void prune( uint i, Real lb );
    void print( FILE* is, uint n ) const;
};

const Real Chart::EPSILON = 10e-3;


Chart::Chart() {

    cell_  = NULL;
}


Chart::~Chart() {

    delete[] cell_;
}


inline void Chart::set_size( uint L, uint D ) {

    delete[] cell_;
    cell_  = new Cell[MAX_LENGTH*L];
    first_.resize( MAX_LENGTH );
    last_.resize( MAX_LENGTH );
    L_ = L;
    D_ = D;
}


inline void Chart::viterbi_init( uint n ) {
    
    Cell* c = cell_;
    for ( uint i = 0; i < n; ++i ) {
	for ( Label label = 0; label < L_; ++label ) {
	    c->label = label;
	    if ( label == 0 ) {
		c->prev = NULL;
		first_[i] = c;
	    }else {
		c->prev = c-1;
	    }
	    if ( label == L_-1 ) {
		c->next = NULL;
		last_[i] = c;
	    }else {
		c->next = c+1;
	    }
	    ++c;
	}
    }
}


inline void Chart::staggered_init( uint n ) {

    Cell* c = cell_;
    for ( uint i = 0; i < n; ++i ) {
	// One active label
	first_[i] = c;
	c->label  = 0;
	c->next   = c + 1;
	c->prev   = NULL;
	++c;
	
	// One degenerate label
	last_[i] = c;
	c->label = L_;
	c->next  = NULL;
	c->prev  = c - 1;
	c += (L_-1);
    }
}


inline void Chart::prune( uint i, Real lb ) {

    for ( Cell* c = first_[i]; c != NULL; c = c->next ) {
	if ( c->alpha + c->beta + EPSILON < lb ) {
	    Cell* c_prev = c->prev;
	    Cell* c_next = c->next;
	    if ( c_prev == NULL ) {
		if ( c_next == NULL ) {
		    fprintf( stderr, HERE "An error occurred in pruning\n" );
		    exit( -1 );
		}
		c_next->prev = NULL;
		first_[i] = c_next;
	    }else if ( c_next == NULL ) {
		c_prev->next = NULL;
		last_[i] = c_prev;
	    }else {
		c_prev->next = c_next;
		c_next->prev = c_prev;
	    }
	}
    }
}


inline void Chart::expand( uint i ) {

    // Memorize the last cell (= degenerate label)
    Cell* last  = last_[i];
    Label label = last->label;
    Real  alpha = last->alpha;
    Real  beta  = last->beta;
    Cell* prev  = last->prev;
	
    // Expand active labels
    if ( label+1 < L_+D_ ) {
	Cell* c = last;
	Label y_first = pow2( label-L_ );
	Label y_last = 2*y_first;
	for ( Label y = y_first; y < y_last; ++y ) {
	    c->label = y;
	    c->alpha = alpha; // + node_score( i, y ) - node_score( i, label );
	    c->beta  = beta;
	    if ( y == y_first ) {
		c->prev = prev;
	    }else {
		c->prev = c - 1;
	    }
	    c->next = c + 1;
	    ++c;
	}
	
	// degenerate label
	c->label = label + 1;
	c->alpha = alpha; // + node_score( i, c->label ) - node_score( i, label );
	c->beta  = beta;
	c->next  = NULL;
	c->prev  = c - 1;
	last_[i] = c;
    }else {
	Cell* c = last;
	Label y_first = pow2( label-L_ );
	for ( Label y = y_first; y < L_; ++y ) {
	    c->label = y;
	    c->alpha = alpha; // + node_score( i, y ) - node_score( i, label );
	    c->beta  = beta;
	    if ( y == y_first ) {
		c->prev = prev;
	    }else {
		c->prev = c - 1;
	    }
	    if ( y == L_ - 1 ) {
		c->next = NULL;
		last_[i] = c;
	    }else {
		c->next = c + 1;
		++c;
	    }
	}
    }
}


inline Cell* Chart::first( uint i ) {
    
    return first_[i];
}


inline const Cell* Chart::first( uint i ) const {
    
    return first_[i];
}


inline Cell* Chart::last( uint i ) {
    
    return last_[i];
}


inline const Cell* Chart::last( uint i ) const {
    
    return last_[i];
}


inline void Chart::print( FILE* is, uint n ) const {

    for ( uint i = 0; i < n; ++i ) {
	for ( Cell* c = first_[i]; c != NULL; c = c->next ) {
	    fprintf( is, "%d:%f\t", c->label, c->alpha );
	}
	fprintf( is, "\n" );
    }
}


}

#endif
