#ifndef __DOUBLEARRAY__
#define __DOUBLEARRAY__

#include <cstring>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <set>
#include "util.h"
#include "timer.h"

/*
 * Simple static doulbe array
 */
namespace SD {

class DoubleArray {
  private:    
    static const char ENDMARKER;
    static const char DUMMYCHAR;
    static const char NUMERIC;
    static const uint ROOT;

    // the main data structure
    uint              array_size_;
    std::vector<uint> base_;
    std::vector<char> check_;
        
    // key and key_id
    uint                     key_num_;
    std::vector<std::string> key_;
    std::vector<uint>        key_id_;    // mapping from base value to key ID
    
  public:
    DoubleArray() {};
    ~DoubleArray() {};
    DoubleArray( const DoubleArray& obj );
    DoubleArray& operator=( const DoubleArray& obj );

    int build( const std::vector<std::string>& key );
    int build_reverse( const std::vector<std::string>& key );
    int save( FILE* os ) const;
    int load( FILE* is );

    uint exact_match( const char* s ) const;
    uint exact_match( const char* first, const char* last ) const;
    uint longest_match( const char* first, const char* last ) const;
    uint longest_match_reverse( const char* first, const char* last ) const;
    uint key_num() const;
    
  private:
    void init_array();
    void build( uint n, const std::string& prefix, const std::vector<std::string>& v, uint& min_base, std::set<uint>& reserved_id );
    int store_key( const std::vector<std::string>& key );
    void decide_base( uint n, uint min_base, const std::set<char>& c, const std::set<uint>& reserved_id );
};


// Use control characters as the special characters.
const char DoubleArray::ENDMARKER = '\x00';
const char DoubleArray::DUMMYCHAR = '\x01';
const char DoubleArray::NUMERIC   = '#';
const uint DoubleArray::ROOT      = 1;


/**********************************************************************
 *
 *  NOTE:
 *
 *  Every node ID (including the root) is equal or larger than 1.
 *
 *  const ROOT(=1) is the ID of the root node.
 *
 *  Min value of base is 1. This gurantees that base_[i] + c => 1
 *  for all i and c.
 *
 *  check_[i] = DUMMYCHAR implies that node ID i does not exist.
 *
 **********************************************************************/


int DoubleArray::build_reverse( const std::vector<std::string>& key ) {

    std::vector<std::string> key2;
    std::copy( key.begin(), key.end(), std::back_inserter( key2 ) );
    for ( uint i = 0; i < key2.size(); i++ ) {
	std::reverse( key2[i].begin(), key2[i].end() );
    }
    return build( key2 );
}


int DoubleArray::build( const std::vector<std::string>& key ) {
    
    init_array();
    
    // append ENDMARKER to the keys.
    std::vector<std::string> v;
    v.resize( key.size() );
    copy( key.begin(), key.end(), v.begin() );
    for ( uint i = 0; i < key.size(); i++ ) {
	v[i].append( 1, ENDMARKER );
    }
    sort( v.begin(), v.end() );

    // the main part
    uint min_base = 1;
    std::set<uint> reserved_id;
    reserved_id.insert( ROOT );
    build( ROOT, "", v, min_base, reserved_id );

    if ( store_key( key ) == -1 ) return -1;
    
    return 0;
}


void DoubleArray::init_array() {
    
    array_size_ = 2;
    
    base_.resize( array_size_ );
    base_[0] = 0;
    base_[1] = 1;
    
    check_.resize( array_size_ );
    check_[0] = DUMMYCHAR;
    check_[1] = DUMMYCHAR;
}


int DoubleArray::store_key( const std::vector<std::string>& key ) {

    key_num_ = key.size();
    key_.resize( key_num_ );
    copy( key.begin(), key.end(), key_.begin() );

    key_id_.resize( array_size_, key_num_ );
    for ( uint i = 0; i < key_num_; ++i ) {
	uint node = ROOT;
	for ( const char* s = key_[i].c_str(); *s != '\0'; ++s ) {
	    uint next = base_[node] + static_cast<uint>(static_cast<uchar>(*s));
	    if ( array_size_ < next || check_[next] != *s ) {
		fprintf( stderr, HERE "unexpected error\n" );
		return -1;
	    }
	    node = next;
	}
	uint next = base_[node] + static_cast<uint>(static_cast<uchar>(ENDMARKER));
	if ( array_size_ < next || check_[next] != ENDMARKER ) {
	    fprintf( stderr, HERE "unexpected error\n" );
	    return -1;
	}
	key_id_[next] = i;
    }
    
    return 0;
}


void DoubleArray::build( uint n, const std::string& prefix, const std::vector<std::string>& v, uint& min_base, std::set<uint>& reserved_id ) {

    // prefix ³Ƥʸ󤹤
    std::set<char> successor;
    std::vector<std::string>::const_iterator lb = lower_bound( v.begin(), v.end(), prefix );
    while ( lb != v.end() && lb->find( prefix, 0 ) == 0 ) {
	if ( *lb != prefix ) {
	    successor.insert( lb->at( prefix.length() ) );
	}
	++lb;
    }
    if ( successor.size() == 0 ) return;

    // decide base_[n]
    decide_base( n, min_base, successor, reserved_id );
	
    //  arrange the array size
    uint max = 0;
    for ( std::set<char>::iterator it = successor.begin(); it != successor.end(); ++it ) {
	if ( max < base_[n] + static_cast<uint>(static_cast<uchar>(*it)) ) {
	    max = base_[n] + static_cast<uint>(static_cast<uchar>(*it));
	}
    }
    if ( array_size_ < max + 1 ) {
	base_.resize( max+1 );
	check_.resize( max+1, DUMMYCHAR );
	array_size_ = max+1;
    }
    
    //  set the check_
    for ( std::set<char>::iterator it = successor.begin(); it != successor.end(); ++it ) {
	check_[base_[n] + static_cast<uint>(static_cast<uchar>(*it))] = *it;
    }
    
    //  update min_base and reserved_id
    min_base = base_[n] + 1;
    for ( std::set<char>::iterator it = successor.begin(); it != successor.end(); ++it ) {
	reserved_id.insert( base_[n] + static_cast<uint>(static_cast<uchar>(*it)) );
    }
    
    //  process recursively
    for ( std::set<char>::iterator it = successor.begin(); it != successor.end(); ++it ) {
	build( base_[n] + static_cast<uint>(static_cast<uchar>(*it)), prefix + *it, v, min_base, reserved_id );
    }
}


void DoubleArray::decide_base( uint n, uint min_base, const std::set<char>& c, const std::set<uint>& reserved_id ) {

    base_[n] = min_base;
    bool collision = false;
    do {
	// check collision
	collision = false;
	for ( std::set<char>::const_iterator it = c.begin(); it != c.end(); ++it ) {
	    if ( reserved_id.find( base_[n] + static_cast<uint>(static_cast<uchar>(*it)) ) != reserved_id.end() ) {
		collision = true;
		break;
	    }
	}
	
	//  ͤʤʤޤ array_[n].base ͤ䤷Ƥ
	if ( collision == true ) {
	    base_[n]++;
	}
    }while( collision == true );
}


int DoubleArray::save( FILE* os ) const {

    if ( fwrite( &array_size_, sizeof( uint ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save double array\n" );
	return -1;
    }
    if ( fwrite( &key_num_, sizeof( uint ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save double array\n" );
	return -1;
    }
    if ( fwrite( &base_[0], sizeof( uint ), array_size_, os ) != array_size_ ) {
	fprintf( stderr, HERE "failed to save double array\n" );
	return -1;
    }
    if ( fwrite( &check_[0], sizeof( char ), array_size_, os ) != array_size_ ) {
	fprintf( stderr, HERE "failed to save double array\n" );
	return -1;
    }
    if ( fwrite( &key_id_[0], sizeof( uint ), array_size_, os ) != array_size_ ) {
	fprintf( stderr, HERE "failed to save double array\n" );
	return -1;
    }
    for ( uint i = 0; i < key_num_; ++i ) {
	uint size = key_[i].length() + 1; // nullʸޤ
	if ( fwrite( &size, sizeof( uint ), 1, os ) != 1 ) {
	    fprintf( stderr, HERE "failed to save double array\n" );
	    return -1;
	}
	if ( fwrite( key_[i].c_str(), sizeof( char ), size, os ) != size ) {
	    fprintf( stderr, HERE "failed to save double array\n" );
	    return -1;
	}
    }
    return 0;
}


int DoubleArray::load( FILE* is ) {

    if ( fread( &array_size_, sizeof( uint ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load double array\n" );
	return -1;
    }
    if ( fread( &key_num_, sizeof( uint ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load double array\n" );
	return -1;
    }
    base_.resize( array_size_ );
    if ( fread( &base_[0], sizeof( uint ), array_size_, is ) != array_size_ ) {
	fprintf( stderr, HERE "failed to load double array\n" );
	return -1;
    }
    check_.resize( array_size_ );
    if ( fread( &check_[0], sizeof( char ), array_size_, is ) != array_size_ ) {
	fprintf( stderr, HERE "failed to load double array\n" );
	return -1;
    }
    key_id_.resize( array_size_ );
    if ( fread( &key_id_[0], sizeof( uint ), array_size_, is ) != array_size_ ) {
	fprintf( stderr, HERE "failed to load double array\n" );
	return -1;
    }
    key_.resize( key_num_ );
    for ( uint i = 0; i < key_num_; ++i ) {
	uint size;
	if ( fread( &size, sizeof( uint ), 1, is ) != 1 ) {
	    fprintf( stderr, HERE "failed to load double array\n" );
	    return -1;
	}
	char buff[size];
	if ( fread( buff, sizeof( char ), size, is ) != size ) {
	    fprintf( stderr, HERE "failed to load double array\n" );
	    return -1;
	}
	key_[i].assign( buff );
    }
    
    return 0;
}


uint DoubleArray::exact_match( const char* s ) const {

    return exact_match( s, s + strlen(s) );
}


uint DoubleArray::exact_match( const char* first, const char* last ) const {

    bool numeric = false;
    uint node = ROOT;
    for ( const char* s = first; s != last; ++s ) {
	if ( '0' <= *s && *s <= '9' ) {
	    if ( numeric == true ) {
		continue;
	    }else {
		numeric = true;
		uint next_node = base_[node] + static_cast<uint>(static_cast<uchar>(NUMERIC));
		if ( next_node < array_size_ && check_[next_node] == NUMERIC ) {
		    node = next_node;	    
		}else {
		    return key_num_;
		}
	    }
	}else {
	    numeric = false;
	    uint next_node = base_[node] + static_cast<uint>(static_cast<uchar>(*s));
	    if ( next_node < array_size_ && check_[next_node] == *s ) {
		node = next_node;	    
	    }else {
		return key_num_;
	    }
	}
    }
    
    uint next = base_[node] + static_cast<uint>(static_cast<uchar>(ENDMARKER));
    if ( next < array_size_ && check_[next] == ENDMARKER ) {
	return key_id_[next];
    }else {
	return key_num_;
    }
}


uint DoubleArray::longest_match( const char* first, const char* last ) const {

    uint node    = ROOT;
    uint ret     = key_num_;
    for ( const char* s = first; s != last; ++s ) {
	uint next = base_[node] + static_cast<uint>(static_cast<uchar>(*s));
	if ( next < array_size_ && check_[next] == *s ) {
	    node = next;
	}else {
	    break;
	}
	next = base_[node] + static_cast<uint>(static_cast<uchar>(ENDMARKER));
	if ( next < array_size_ && check_[next] == ENDMARKER ) {
	    ret = key_id_[next];
	}
    }
    return ret;
}


uint DoubleArray::longest_match_reverse( const char* first, const char* last ) const {

    uint node    = ROOT;
    uint ret     = key_num_;
    for ( const char* s = last-1; ; --s ) {
	uint next = base_[node] + static_cast<uint>(static_cast<uchar>(*s));
	if ( next < array_size_ && check_[next] == *s ) {
	    node = next;
	}else {
	    break;
	}
	next = base_[node] + static_cast<uint>(static_cast<uchar>(ENDMARKER));
	if ( next < array_size_ && check_[next] == ENDMARKER ) {
	    ret = key_id_[next];
	}
	if ( s == first ) break;
    }
    return ret;
}
    


uint DoubleArray::key_num() const {
    
    return key_num_;
}


}
#endif
