#ifndef __MODEL__
#define __MODEL__

#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>

#include "util.h"
#include "dense_matrix.h"
#include "timer.h"

namespace SD {

class Model {
  private:
    uint L_;
    uint D_;
    uint M_;
    uint V_;
    TimerPool pool_;
    Timer*    timer_;
    DenseMatrix<Real> word_score_;
    DenseMatrix<Real> prev_word_score_;
    DenseMatrix<Real> next_word_score_;
    DenseMatrix<Real> prefix_score_;
    DenseMatrix<Real> suffix_score_;
    DenseMatrix<Real> cap_score_;
    DenseMatrix<Real> node_score_;
    DenseMatrix<Real> edge_score_;
  public:
    Model();
    Model( const Model& obj );
    ~Model();
    Model& operator=( const Model& obj );
    int read( FILE* is );
    int load( FILE* is );
    int save( FILE* os ) const;
    uint L() const;
    uint D() const;
    void cache_node_score( const Token* x, uint n );
    Real node_score( uint i, Label y ) const;
    Real edge_score( Label y1, Label y2 ) const;
};


Model::Model() {

    //timer_ = pool_.push( "cache" );
}

Model::~Model() {
    
    //pool_.print();
}

int Model::read( FILE* is ) {

    DenseMatrix<Real> tmp;
    if ( word_score_.read( is ) == -1 ) return -1;
    if ( prev_word_score_.read( is ) == -1 ) return -1;
    if ( next_word_score_.read( is ) == -1 ) return -1;
    if ( prefix_score_.read( is ) == -1 ) return -1;
    if ( suffix_score_.read( is ) == -1 ) return -1;
    if ( cap_score_.read( is ) == -1 ) return -1;
    if ( tmp.read( is ) == -1 ) return -1;
    
    L_ = word_score_.column_size();
    D_ = degenerate_label_num(L_);
    M_ = pow2( D_-1 );
    V_ = word_score_.row_size();
    
    edge_score_.init( L_+D_, L_+D_, ZERO );
    for ( Label y1 = 0; y1 < L_; ++y1 ) {
	for ( Label y2 = 0; y2 < L_; ++y2 ) {
	    edge_score_[y1][y2] = tmp[y1][y2];
	}
    }
    for ( Label y = 0; y < L_; ++y ) {
	for ( Label z = L_; z < L_+D_; ++z ) {
	    edge_score_[y][z] = edge_score_.max( y, y+1, pow2(z-L_), L_ );
	    edge_score_[z][y] = edge_score_.max( pow2(z-L_), L_, y, y+1 );
	}
    }
    for ( Label z1 = L_; z1 < L_+D_; ++z1 ) {
	for ( Label z2 = L_; z2 < L_+D_; ++z2 ) {
	    edge_score_[z1][z2] = edge_score_.max( pow2(z1-L_), L_, pow2(z2-L_), L_ );
	}
    }
    
    return 0;
}


int Model::load( FILE* is ) {

    if ( fread( &L_, sizeof( uint ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load data\n" );
	return -1;
    }
    if ( fread( &D_, sizeof( uint ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load data\n" );
	return -1;
    }
    if ( fread( &M_, sizeof( uint ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load data\n" );
	return -1;
    }
    if ( fread( &V_, sizeof( uint ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load data\n" );
	return -1;
    }
    
    if ( word_score_.load( is ) == -1 ) return -1;
    if ( prev_word_score_.load( is ) == -1 ) return -1;
    if ( next_word_score_.load( is ) == -1 ) return -1;
    if ( prefix_score_.load( is ) == -1 ) return -1;
    if ( suffix_score_.load( is ) == -1 ) return -1;
    if ( cap_score_.load( is ) == -1 ) return -1;
    if ( edge_score_.load( is ) == -1 ) return -1;
    node_score_.init( MAX_LENGTH, L_+D_, ZERO );
    
    return 0;
}


int Model::save( FILE* os ) const {

    if ( fwrite( &L_, sizeof( uint ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save data\n" );
	return -1;
    }
    if ( fwrite( &D_, sizeof( uint ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save data\n" );
	return -1;
    }
    if ( fwrite( &M_, sizeof( uint ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save data\n" );
	return -1;
    }
    if ( fwrite( &V_, sizeof( uint ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save data\n" );
	return -1;
    }
    if ( word_score_.save( os ) == -1 ) return -1;
    if ( prev_word_score_.save( os ) == -1 ) return -1;
    if ( next_word_score_.save( os ) == -1 ) return -1;
    if ( prefix_score_.save( os ) == -1 ) return -1;
    if ( suffix_score_.save( os ) == -1 ) return -1;
    if ( cap_score_.save( os ) == -1 ) return -1;
    if ( edge_score_.save( os ) == -1 ) return -1;
    
    return 0;
}


inline void Model::cache_node_score( const Token* x, uint n ) {
    
    //timer_->startTimer();
    for ( uint i = 0; i < n; ++i ) {
	if ( x[i].id < V_-1 ) {
	    Real* target = node_score_[i];
	    Real* source = word_score_[x[i].id];
	    for ( Label y = 0; y < L_; ++y ) {
		*target = *source;
		++target;
		++source;
	    }
	}else {
	    Real* target  = node_score_[i];
	    Real* source1 = prefix_score_[x[i].prefix_id];
	    Real* source2 = suffix_score_[x[i].suffix_id];
	    Real* source3 = cap_score_[x[i].capitalization];
	    for ( Label y = 0; y < L_; ++y ) {
		*target = *source1 + *source2 + *source3;
		++target;
		++source1;
		++source2;
		++source3;
	    }
	}
    }

    Real* target = node_score_[0];
    Real* source = prev_word_score_[BOS_ID];
    for ( Label y = 0; y < L_; ++y ) {
	*target += *source;
	++target;
	++source;
    }
    for ( uint i = 1; i < n; ++i ) {
	target = node_score_[i];
	source = prev_word_score_[x[i-1].id];
	for ( Label y = 0; y < L_; ++y ) {
	    *target += *source;
	    ++target;
	    ++source;
	}
    }

    for ( uint i = 0; i < n-1; ++i ) {
	target = node_score_[i];
	source = next_word_score_[x[i+1].id];
	for ( Label y = 0; y < L_; ++y ) {
	    *target += *source;
	    ++target;
	    ++source;
	}	
    }
    target = node_score_[n-1];
    source = next_word_score_[EOS_ID];
    for ( Label y = 0; y < L_; ++y ) {
	*target += *source;
	++target;
	++source;
    }
    
    for ( uint i = 0; i < n; ++i ) {
	estimate_ub( node_score_[i], L_, D_, M_ );
    }
    //timer_->stopTimer();
}


inline Real Model::node_score( uint i, Label y ) const {

    return node_score_[i][y];
}


inline Real Model::edge_score( Label y1, Label y2 ) const {

    return edge_score_[y1][y2];
}


inline uint Model::L() const {

    return L_;
}

inline uint Model::D() const {

    return D_;
}


}

#endif
