#include <iostream>
#include <fstream>

#include "util.h"
#include "token.h"
#include "timer.h"
#include "string_vector.h"
#include "encoder.h"
#include "tagger.h"


using namespace SD;


int load( const char* modelfile, Encoder& encoder, Tagger& tagger, StringVector& labelset ) {

    FILE* is = fopen( modelfile, "rb" );
    if ( is == NULL ) {
	fprintf( stderr, HERE "cannot open %s\n", modelfile );
	return -1;
    }
    if ( encoder.load( is ) == -1 ) return -1;
    if ( labelset.load( is ) == -1 ) return -1;
    if ( tagger.load( is ) == -1 ) return -1;
    fclose( is );

    return 0;
}


inline void init( const char* s, Token* x, size_t& n ) {

    n = 0;
    for ( ; *s != '\0'; ++s ) {
	x[n].begin = s;
	for ( ; *s != SEP && *s != '\0'; ++s ) {
	    ;
	}
	x[n].end = s;
	n++;
    }
    if ( 1 <= n && x[n-1].begin != x[n-1].end-1 && *(x[n-1].end-1) == '.' ) {
	x[n-1].end--;
	++n;
	x[n-1].begin = x[n-2].end;
	x[n-1].end   = x[n-2].end+1;
    }
}


inline int print_result( FILE* os, const Token* x, const Label* y, SD::uint n, const StringVector& label_set ) {

    for ( SD::uint i = 0; i < n; ++i ) {
	if ( i != 0 ) {
	    fprintf( os, " " );
	}
	for ( const char* s = x[i].begin; s != x[i].end; ++s ) {
	    fprintf( os, "%c", *s );
	}
	fprintf( os, "/" );
	fprintf( os, "%s", label_set[y[i]] );
    }
    fprintf( os, "\n" );
    
    return 0;
}


void usage( const char* cmd ) {

    fprintf( stderr, "%s [-d 0(Viterib)|1(Staggered, default)] [-v] <model>\n", cmd );
}


int get_option( SD::uint argc, const char** argv, int& verbose, int& algorithm ) {

    for ( SD::uint i = 1; i < argc-1; ) {
	if ( strcmp( argv[i], "-v" ) == 0 ) {
	    verbose = 1;
	    ++i;
	}else if ( strcmp( argv[i], "-d" ) == 0 ) {
	    algorithm = atoi( argv[i+1] );
	    i += 2;
	}else {
	    return -1;
	}
    }
    return 0;
}


int main( int argc, const char** argv ) {

    if ( argc < 2 ) {
	usage( argv[0] );
	return -1;
    }
    int verbose   = 0;
    int algorithm = 1;
    if ( get_option( argc, argv, verbose, algorithm ) == -1 ) return -1;
    const char* modelfile = argv[argc-1];
    
    FILE* is = stdin;
    FILE* os = stdout;

    TimerPool pool;
    Timer* timer = pool.push( "run time" );
    
    Encoder  encoder;
    Tagger   tagger;
    StringVector label_set;
    if ( load( modelfile, encoder, tagger, label_set ) == -1 ) return -1;
    
    size_t n;
    char*  s = new char[BUFF_SIZE];
    Token* x = new Token[MAX_LENGTH];
    Label* y = new Label[MAX_LENGTH];

    timer->startTimer();
    while ( fgets( s, BUFF_SIZE, is ) != NULL ) {
	if ( s[strlen(s)-1] != '\n' ) {
	    fprintf( stderr, HERE "too long input!\n" );
	    continue;
	}
	s[strlen(s)-1] = '\0';

	init( s, x, n );
	encoder.run( x, n );
	if ( algorithm == 0 ) {
	    tagger.viterbi( x, y, n );
	}else {
	    tagger.staggered( x, y, n );
	}
	print_result( os, x, y, n, label_set );
    }
    timer->stopTimer();
    if ( verbose == 1 ) pool.print();
    
    delete[] s;
    delete[] x;
    delete[] y;
    
    return 0;
}
