#ifndef __TAGGER__
#define __TAGGER__

#include <iostream>
#include <fstream>
#include <vector>

#include "util.h"
#include "token.h"
#include "model.h"
#include "chart.h"
#include "timer.h"

namespace SD {

class Tagger {
  private:
    uint L_;
    uint D_;
    Real  lb_;
    Chart chart_;
    Model model_;
    TimerPool pool_;
    Timer* timer_;
  public:
    Tagger();
    Tagger( const Tagger& obj );
    ~Tagger();
    Tagger& operator=( const Tagger& obj );
    int load( FILE* is );
    void viterbi( const Token* x, Label* y, uint n );
    void staggered( const Token* x, Label* y, uint n );
    Real det( const Token* x, Label* y, uint n );
    Real det2( const Token* x, Label* y, uint n );
  private:
    void fw_staggered( Label* y, uint n, bool flag );
    void bw_staggered( Label* y, uint n );
    Real traverse_backlink1( Label* y, uint n ) const;
    Real traverse_backlink2( Label* y, uint n ) const;
};


Tagger::Tagger() {

    //timer_ = pool_.push( "tagger" );
}


Tagger::~Tagger() {

    //pool_.print();
}


int Tagger::load( FILE* is ) {
    
    if ( model_.load( is ) == -1 ) return -1;
    chart_.set_size( model_.L(), model_.D() );
    L_ = model_.L();
    D_ = model_.D();
    
    return 0;
}


inline void Tagger::viterbi( const Token* x, Label* y, uint n ) {

    if ( n == 0 ) return;

    chart_.viterbi_init( n );
    model_.cache_node_score( x, n );

    for ( Cell* c = chart_.first(0); c != NULL; c = c->next ) {
	c->alpha    = model_.node_score( 0, c->label );
	c->backlink = NULL;
    }
    for ( uint i = 1; i < n; ++i ) {
	Cell* c1 = chart_.first(i-1);
	for ( Cell* c2 = chart_.first(i); c2 != NULL; c2 = c2->next ) {
	    c2->alpha    = c1->alpha + model_.edge_score( c1->label, c2->label ) + model_.node_score( i, c2->label );
	    c2->backlink = c1;
	}
	for ( c1 = c1->next; c1 != NULL; c1 = c1->next ) {
	    for ( Cell* c2 = chart_.first(i); c2 != NULL; c2 = c2->next ) {
		Real tmp_alpha = c1->alpha + model_.edge_score( c1->label, c2->label ) + model_.node_score( i, c2->label );
		if ( c2->alpha < tmp_alpha ) {
		    c2->alpha    = tmp_alpha;
		    c2->backlink = c1;
		}
	    }
	}
    }
    
    traverse_backlink1( y, n );
}


inline void Tagger::staggered( const Token* x, Label* y, uint n ) {

    if ( n == 0 ) return;

    chart_.staggered_init( n );
    lb_ = det( x, y, n );
    for ( uint iter_num = 0; ; ++iter_num ) {
	if ( iter_num%2 == 0 ) {
	    fw_staggered( y, n, (iter_num!=0) );
	}else {
	    bw_staggered( y, n );
	}
	bool is_optimal = true;
	for ( uint i = 0; i < n; ++i ) {
	    if ( L_ <= y[i] ) {
		is_optimal = false;
		break;
	    }
	}
	if ( is_optimal ) {
	    return;
	}else {
	    for ( uint i = 0; i < n; ++i ) {
		if ( L_ <= y[i] ) chart_.expand( i );
	    }
	}
    }
}


inline void Tagger::fw_staggered( Label* y, uint n, bool prune_flag ) {

    for ( Cell* c = chart_.first(0); c != NULL; c = c->next ) {
	c->alpha    = model_.node_score( 0, c->label );
	c->lb       = c->alpha;
	c->backlink = NULL;
    }
    if ( prune_flag ) chart_.prune( 0, lb_ );
    
    for ( uint i = 1; i < n; ++i ) {
	Cell* c1 = chart_.first(i-1);
	if ( c1->label < L_ ) {
	    for ( Cell* c2 = chart_.first(i); c2 != NULL; c2 = c2->next ) {
		Real delta  = model_.edge_score( c1->label, c2->label ) + model_.node_score( i, c2->label );
		c2->alpha    = c1->alpha + delta;
		c2->lb       = c1->lb + delta;
		c2->backlink = c1;
	    }
	}else {
	    for ( Cell* c2 = chart_.first(i); c2 != NULL; c2 = c2->next ) {
		c2->alpha    = c1->alpha + model_.edge_score( c1->label, c2->label ) + model_.node_score( i, c2->label );
		c2->lb       = MIN_REAL;
		c2->backlink = c1;
	    }
	}
	for ( c1 = c1->next; c1 != NULL; c1 = c1->next ) {
	    if ( c1->label < L_ ) {
		for ( Cell* c2 = chart_.first(i); c2 != NULL; c2 = c2->next ) {
		    Real delta = model_.edge_score( c1->label, c2->label ) + model_.node_score( i, c2->label );
		    Real score = c1->alpha + delta;
		    Real lb    = c1->lb + delta;
		    if ( c2->alpha < score ) {
			c2->alpha    = score;
			c2->backlink = c1;
		    }
		    if ( c2->lb < lb ) {
			c2->lb = lb;
		    }
		}
	    }else {
		for ( Cell* c2 = chart_.first(i); c2 != NULL; c2 = c2->next ) {
		    Real score = c1->alpha + model_.edge_score( c1->label, c2->label ) + model_.node_score( i, c2->label );
		    if ( c2->alpha < score ) {
			c2->alpha    = score;
			c2->backlink = c1;
		    }
		}
	    }
	}
	
	if ( prune_flag ) chart_.prune( i, lb_ );
    }

    // Update lower bound
    if ( prune_flag ) {
	for ( const Cell* c = chart_.first(n-1); c != NULL; c = c->next ) {
	    if ( c->label < L_ && lb_ < c->lb ) {
		lb_ = c->lb;
	    }
	}
    }
    traverse_backlink1( y, n );
}



inline void Tagger::bw_staggered(  Label* y, uint n ) {

    for ( Cell* c = chart_.first(n-1); c != NULL; c = c->next ) {
	c->beta     = ZERO;
	c->lb       = ZERO;
	c->backlink = NULL;
    }
    chart_.prune( n-1, lb_ );

    for ( uint i = n-1; 1 <= i; --i ) {
	Cell* c2 = chart_.first(i);
	if( c2->label < L_ ) {
	    for ( Cell* c1 = chart_.first(i-1); c1 != NULL; c1 = c1->next ) {
		Real delta = model_.edge_score( c1->label, c2->label ) + model_.node_score( i, c2->label );
		c1->beta     = c2->beta + delta;
		c1->lb       = c2->lb + delta;
		c1->backlink = c2;
	    }
	}else {
	    for ( Cell* c1 = chart_.first(i-1); c1 != NULL; c1 = c1->next ) {
		c1->beta     = c2->beta + model_.edge_score( c1->label, c2->label ) + model_.node_score( i, c2->label );
		c1->lb       = MIN_REAL;
		c1->backlink = c2;
	    }
	}
	for ( c2 = c2->next; c2 != NULL; c2 = c2->next ) {
	    if ( c2->label < L_ ) {
		for ( Cell* c1 = chart_.first(i-1); c1 != NULL; c1 = c1->next ) {
		    Real delta = model_.edge_score( c1->label, c2->label ) + model_.node_score( i, c2->label );
		    Real score = delta + c2->beta;
		    Real lb    = delta + c2->lb;
		    if ( c1->beta < score ) {
			c1->beta     = score;
			c1->backlink = c2;
		    }
		    if ( c1->lb < lb ) {
			c1->lb = lb;
		    }
		}
	    }else {
		for ( Cell* c1 = chart_.first(i-1); c1 != NULL; c1 = c1->next ) {
		    Real score = model_.edge_score( c1->label, c2->label ) + model_.node_score( i, c2->label ) + c2->beta;
		    if ( c1->beta < score ) {
			c1->beta    = score;
			c1->backlink = c2;
		    }
		}
	    }
	}
	chart_.prune(i-1,lb_);
    }

    // Update lower bound
    for ( const Cell* c = chart_.first(0); c != NULL; c = c->next ) {
	if ( c->label < L_ ) {
	    Real score = model_.node_score( 0, c->label ) + c->lb;
	    if ( lb_ < score ) {
		lb_ = score;
	    }
	}
    }
    
    traverse_backlink2( y, n );
}



inline Real Tagger::det( const Token* x, Label* y, uint n ) {

    if ( n == 0 ) return ZERO;

    model_.cache_node_score( x, n );
    
    Real best_score = MIN_REAL;
    for ( Label label = 0; label < L_; ++label ) {
	Real score = model_.node_score( 0, label );
	if ( best_score < score ) {
	    y[0] = label;
	    best_score = score;
	}
    }
    for ( uint i = 1; i < n; ++i ) {
	Real best_delta = MIN_REAL;
	for ( Label label = 0; label < L_; ++label ) {
	    Real delta = model_.edge_score( y[i-1], label ) + model_.node_score( i, label );
	    if ( best_delta < delta ) {
		y[i] = label;
		best_delta = delta;
	    }
	}
	best_score += best_delta;
    }
    
    return best_score;
}


inline Real Tagger::det2( const Token* x, Label* y, uint n ) {

    if ( n == 0 ) return ZERO;

    model_.cache_node_score( x, n );
    
    Real best_score = MIN_REAL;
    Label first_label = 0;
    Label last_label  = 1;
    for ( uint iter_num = 0; iter_num < D_+1; ++iter_num ) {
	for ( Label label = first_label; label < last_label; ++label ) {
	    Real score = model_.node_score( 0, label );
	    if ( best_score < score ) {
		y[0] = label;
		best_score = score;
	    }
	}
	if ( iter_num == D_ || best_score > model_.node_score( 0, L_+iter_num ) ) break;
	first_label = last_label;
	last_label  = std::min( L_, first_label*2 );
    }
    
    for ( uint i = 1; i < n; ++i ) {
	Real  best_delta = MIN_REAL;
	Label first_label = 0;
	Label last_label  = 1;
	for ( uint iter_num = 0; iter_num < D_+1; ++iter_num ) {
	    for ( Label label = first_label; label < last_label; ++label ) {
		Real delta = model_.edge_score( y[i-1], label ) + model_.node_score( i, label );
		if ( best_delta < delta ) {
		    y[i] = label;
		    best_delta = delta;
		}
	    }
	    if ( iter_num == D_ || best_delta > model_.edge_score( y[i-1], L_+iter_num ) + model_.node_score( i, L_+iter_num ) ) break;
	    first_label = last_label;
	    last_label  = std::min( L_, first_label*2 );
	}
	best_score += best_delta;
    }
    
    return best_score;
}



inline Real Tagger::traverse_backlink1( Label* y, uint n ) const {

    const Cell* best_c = NULL;
    Real best_score = MIN_REAL;
    for ( const Cell* c = chart_.first(n-1); c != NULL; c = c->next ) {
	Real score = c->alpha;
	if ( best_score < score ) {
	    best_c     = c;
	    best_score = score;
	}
    }
    
    y[n-1] = best_c->label;
    for ( uint i = n-1; 1 <= i; --i ) {
	best_c = best_c->backlink;
	y[i-1] = best_c->label;
    }
    return best_score;
}


inline Real Tagger::traverse_backlink2( Label* y, uint n ) const {

    const Cell* best_c = NULL;
    Real best_score   = MIN_REAL;
    
    for ( const Cell* c = chart_.first(0); c != NULL; c = c->next ) {
	Real score = model_.node_score( 0, c->label ) + c->beta;
	if ( best_score < score ) {
	    best_c = c;
	    best_score = score;
	}
    }
    
    for ( uint i = 0; i < n; ++i ) {
	y[i]   = best_c->label;
	best_c = best_c->backlink;
    }
    return best_score;
}



}

#endif
