#ifndef TIMER_H
#define TIMER_H

#include <sys/time.h>
#include <sys/resource.h>
#include <string>
#include <iostream>
#include <iomanip>
#include <cstdio>
#include <vector>

#define USE_REAL_CLOCK

class Timer {
private:
  
  unsigned long long _start;
  unsigned long long _end;
  unsigned long long _elapsed;
  
  unsigned int _trial;
  
  const char* _label;
  const char* _unit;
  
public:
  
  static const unsigned long long overhead;
  static const long double clock;
  
  Timer (const char* p = "timer", const char* u = "trial") : _start (0), _end (0), _elapsed (0), _trial (0), _label (p), _unit (u) {};

  void startTimer () { _start = rdtsc (); };
  
  void stopTimer () {
    _elapsed += (rdtsc () - _start - overhead);
    ++_trial;
  };

  static unsigned long long rdtsc () {
    
#if defined(__i386__)
    
    unsigned long long tsc;
    __asm__ __volatile__(".byte 0x0f, 0x31" : "=A" (tsc));
    return (tsc);

#elif (defined(__amd64__) || defined(__x86_64__))

    unsigned long long lo, hi;
    __asm__ __volatile__( "rdtsc" : "=a" (lo), "=d" (hi) );
    return (lo | (hi << 32));

#elif (defined(__powerpc__) || defined(__ppc__))

    unsigned long long tbl, tbu0, tbu1;
    
    do {
        __asm__ __volatile__( "mftbu %0" : "=r" (tbu0) );
        __asm__ __volatile__( "mftb  %0" : "=r" (tbl ) );
        __asm__ __volatile__( "mftbu %0" : "=r" (tbu1) );
    } while (tbu0 != tbu1);
    
    return (tbl);

#elif defined(__sparc__)

    unsigned long long tick;
    __asm__ __volatile__ (".byte 0x83, 0x41, 0x00, 0x00");
    __asm__ __volatile__ ("mov   %%g1, %0" : "=r" (tick));
    return (tick);
  
#elif defined(__alpha__)
  
    unsigned long long cc;
    __asm__ __volatile__ ("rpcc %0" : "=r" (cc));
    return (cc & 0xFFFFFFFF);
  
#elif defined(__ia64__)
  
    unsigned long long itc;
    __asm__ __volatile__ ("mov %0 = ar.itc" : "=r" (itc));
    return (itc);

#endif
  }

  static unsigned long long getOverhead () {
    register unsigned long long now = rdtsc ();
    return (rdtsc () - now);
  }
  
  static long double getCPUClock ();

  void printElapsed () const;
};

class TimerPool {
private:
  typedef std::vector <Timer*> pool_t;
  pool_t _pool;

  // uncopyable
  TimerPool (const TimerPool &s) : _pool (s._pool) {};
  TimerPool& operator= (const TimerPool& b);
public:
  
  TimerPool () : _pool () {};
  Timer* push (const char* p = "timer", const char * u = "trial") {
    Timer* timer_t = new Timer (p, u);
    _pool.push_back (timer_t);
    return timer_t;
  }
  void print () const{
      for (pool_t::const_iterator it = _pool.begin (); it != _pool.end (); ++it)
	  (*it)->printElapsed ();
  }
  ~TimerPool () {
    for (pool_t::iterator it = _pool.begin (); it != _pool.end (); ++it)
      delete *it;
  }
};

#endif /* TIMER_H */
